/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/// Implementation of CBlock and descendents

#define DEBUG_GIF 0
#define DEBUG_COMMENT 0
#define DEBUG_TEXT 0
#define DEBUG_CONTROL 0
#define DEBUG_DRAW 0
#define DEBUG_IMAGE 0
#define DEBUG_COLOR 0     // Use colors.small.gif to keep output down
#define DEBUG_TOGGLE 0
#define DEBUG_EXC 0

#if DEBUG_EXC
void checkHeap(char *title);
void dumpHeap(char *title);
#endif

#define LINESIZE 60

#include <math.h>
#include <Xm/DrawingA.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include "CSGA.h"

// Interface to compression routines
void sgaCompress(FILE *fp, int initCodeSize, unsigned char *pixelsIn,
  int widthIn, int heightIn, int interlaceIn);

// Global variables

static char stringTable[N_STRING_TYPES][LABEL_SIZE]={
    "Global",
    "Image",
    "Control",
    "Plain Text",
    "Application",
    "Comment",
    "Unknown",
};

static char stringDisposalMethod[8][22]={
    "Let Decoder Decide",
    "Leave In Place",
    "Restore To Background",
    "Restore To Previous",
    "Not Defined",
    "Not Defined",
    "Not Defined",
    "Not Defined",
};

/**************************************************************************/
/**************************** CBlock **************************************/
/**************************************************************************/

CBlock::CBlock(CGif *parent, int typeIn) :
    pGif(parent),
    type(typeIn),
    raw(NULL),
    text(NULL)
{
}

CBlock::~CBlock(void)
{
    if(raw) {
	delete [] raw;
	raw = NULL;
    }
    if(text) {
	delete [] text;
	text = NULL;
    }
}

// Stores the raw data in the file for this block and sets the file
// pointer to point to the next block.  After this call, the CBlock
// will use its own data.
int CBlock::storeData(void)
{
    Byte *ptr=pGif->getPtr();
    Byte *rptr, *eof;

  // Get the size of the raw data and store it in rawSize
    storeSize();
    if(rawSize <= 0) {
	errMsg("CBlock::storeData [%s]: There is no data\n",
	  stringType());
	return SGA_ERROR;
    }
    
  // Allocate the space
    raw=new Byte[rawSize];
    if(!raw) {
	errMsg("CBlock::storeData [%s]: Cannot allocate space for data",
	  stringType());
	return SGA_ERROR;
    }

  // Copy the bytes to the block raw data
    rptr=raw;
    eof=rptr+rawSize;
    while(rptr < eof) *rptr++=*ptr++;

  // Set the file pointer to the next block
    pGif->setPtr(pGif->getPtr()+rawSize);
    
    return SGA_OK;
}

// Returns the name of the type as a string.  Does not free the string.
char *CBlock::stringType(void) const
{
    return (type >= 0 && type < N_STRING_TYPES)?
	  stringTable[type]:stringTable[TYPE_UNKNOWN];
}

// Format the data
// "nnnnnnnnnnnn ff ff ff ff ff ff ff ff ff ff nnnnnnnnnn\n"
//  01234567891123456789212345678931234567894123456789512.3
int CBlock::storeText(void)
{
    Byte *ptr;
    Byte ch;
    char *tptr, *hptr, *cptr;
    Byte DataBlockSize;
    char hex[3];
    int lineSize,numberSize;
    const int dataSize=rawSize-introSize;
    int nBlock,nTotal;
    
    if(dataSize <= 0) return SGA_ERROR;

  // Calculate the size of the byte number field
    numberSize=(int)log10((double)dataSize)+2;
    lineSize=42+numberSize;

  // Calculate the text size needed. (Allow an extra line.)
    textSize=(dataSize/10+2)*lineSize+1;

  // Allocate the space
    text=new char[textSize];
    if(!text) {
	errMsg("CBlock::storeText [%s]: Cannot allocate space for text",
	  stringType());
	return SGA_ERROR;
    }

  // Get the bytes from the data blocks
    ptr=raw+introSize;       // Start of data blocks
    tptr=text;               // Start of text
    hptr=tptr+numberSize+1;  // First hex position
    cptr=tptr+numberSize+31; // First character position
    nTotal=0;
    if(*ptr == 0) {
      // This block is a block terminator (e.g. for a Control Block)
	*text='\0';
	return SGA_OK;
    }
    sprintf(tptr,"%*d                                         \n",
      numberSize,nTotal+1);
    while((DataBlockSize=NEXTBYTE)) {
	nBlock=0;
	while(nBlock < DataBlockSize) {
	    ch=NEXTBYTE;
	    if(nTotal && nTotal%10 == 0) {
	      // Start a new line;
		tptr+=lineSize;
		hptr=tptr+numberSize+1;
		cptr=tptr+numberSize+31;
		sprintf(tptr,"%*d                                         \n",
		  numberSize,nTotal+1);
	    }
	    nBlock++; nTotal++;
	    sprintf(hex,"%02X",ch);
	    *hptr++=hex[0];
	    *hptr++=hex[1];
	    *hptr++=' ';
	    if(ch > 31 && ch < 128) {
	      // Is printable ASCII
		*cptr++=ch;
	    } else {
		*cptr++='.';
	    }
	}
    }
#if DEBUG_TEXT
    printf(text);      // May be too big for print
#endif
    return SGA_OK;
}

void CBlock::writeWord(int word,FILE *fp) const
{
  fputc(word&0xff,fp);
  fputc((word/0x100)&0xff,fp);
}

/**************************************************************************/
/**************************** CControl ************************************/
/**************************************************************************/

CControl::CControl(CGif *parent) :
    CBlock(parent,TYPE_CONTROL),
    PackedByte(0),
    TransparentColorFlag(0),
    TransparentIndex(0),
    DelayTime(0),
    DisposalMethod(0)
{
}

CControl::~CControl(void)
{
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CControl::storeSize(void)
{
    Byte *ptr=pGif->getPtr();
    Byte BlockSize,DataBlockSize;

    BlockSize=NEXTBYTE;     /* Should be 4 */
    introSize=BlockSize+1;
    ptr+=BlockSize;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }

    rawSize=(int)(ptr-pGif->getPtr());
}

int CControl::parse(void)
{
    Byte BlockSize;
    Byte ch,*ptr;
    
  // Store the data
    int status=storeData();
    if(status != SGA_OK) return status;

  // Parse the intro block
    ptr=raw;
    BlockSize=NEXTBYTE;     // Should be 0x04=4
  // Parse the Packed byte
  //   Bits 5-7:         Reserved (Ignored)
  //   Bits 2-4:         Disposal Method
  //   Bit 1:            User Input Flag (Ignored)
  //   Bit 0:            Transparent Flag
    PackedByte=ch=NEXTBYTE;
    TransparentColorFlag=ch & MASK0;
    DisposalMethod=(ch & MASK432)>>2;
  // Parse the Delay Time (1/100 sec )
    ch=NEXTBYTE;
    DelayTime=ch + 0x100 * NEXTBYTE;
  // Parse the TransparentIndex
    TransparentIndex=NEXTBYTE;
  // Should only be a block terminator here
#if DEBUG_GIF
    print("    TransparentColorFlag=%d TransparentIndex=%d "
      "DelayTime=%d\n",
      TransparentColorFlag,TransparentIndex,DelayTime);
#endif	

  // Store as text
    if(storeText() != SGA_OK) {
	errMsg("CControl::parse: Cannot store text data",
	  stringType());
	return SGA_ERROR;
    }

    return SGA_OK;
}

int CControl::write(FILE *fp)
{
    Byte b=0;

  // Write extension introducer and extension label
    fputc(EXTENSION,fp);
    fputc(GRAPHCONTROL,fp);

  // Write the block size (always 4)
    fputc(0x04,fp);

  // Write the packed byte
  //   Bits 5-7:         Reserved (Ignored)
  //   Bits 2-4:         Disposal Method
  //   Bit 1:            User Input Flag (Ignored)
  //   Bit 0:            Transparent Flag
  // Only set the disposal method to a non-reserved value
    if(DisposalMethod >= 0 || DisposalMethod < 4) 
      b|=((DisposalMethod<<2)&MASK432);
    if(TransparentColorFlag) b|=MASK0;
    fputc(b,fp);

  // Write the delay time
    writeWord(DelayTime,fp);

  // Write the transparent color index
    fputc(TransparentIndex,fp);

  // Write the block terminator
    fputc(BLOCKTERM,fp);

    return SGA_OK;
}

int CControl::displayData(void)
{
    char line[81];

  // (There is no scrolled window that is managed)

  // Put the intro data in the textArea
    XmTextSetString(textArea,
      "Control Block\n\n");
    sprintf(line,"Transparent Color Flag:    %s\n",
      TransparentColorFlag?"True":"False");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"TransparentIndex:          %d\n",TransparentIndex);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Delay Time:                %d\n",DelayTime);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Disposal Method:           %d [%s]\n",
      DisposalMethod,
      (DisposalMethod >= 0 && DisposalMethod < 8)?
      stringDisposalMethod[DisposalMethod]:"Error");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),"\n");
    
  // There is no data
    XmTextSetCursorPosition(textArea,0);
    XtManageChild(textArea);
      
    return SGA_OK;
}

int CControl::displayOptions(void)
{
    return SGA_OK;
}

/**************************************************************************/
/**************************** CImage **************************************/
/**************************************************************************/

CImage::CImage(CGif *parent) :
    CBlock(parent,TYPE_IMAGE),
    control(NULL),
    raster(NULL),
    imageData(NULL),
    colorData(NULL),
    image(NULL),
    gc(NULL),
    pixmap((Pixmap)NULL),
    prefix(NULL),
    suffix(NULL),
    outCode(NULL)
{
  // Move the Control by taking its pointer
    control=pGif->moveControl();
#if DEBUG_CONTROL
    print("CImage::CImage: Moved Control %x to Image %x\n",
      control,this);
#endif

  // Get the image number
    imageNumber=pGif->getNextImageNumber();

  // Make a graphics context
    XGCValues gcvalues;
    gcvalues.background=WhitePixel(display,screen);
    gcvalues.foreground=BlackPixel(display,screen);
    gc=XCreateGC(display,RootWindow(display,screen),
      GCBackground|GCForeground,&gcvalues);
}

CImage::~CImage(void)
{
  // Delete the Control
#if DEBUG_EXC
    print("~CImage: Before delete control\n");
    checkHeap(" Check Heap:");
#endif
    if(control) {
#if DEBUG_CONTROL
	print("CImage::~CImage: Deleted  Control %x for Image %x\n",
	  control,this);
#endif
	delete control;
	control=NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before delete [] raster\n");
    checkHeap(" Check Heap:");
#endif
  // Delete the raster data
    if(raster) {
	delete [] raster;
	raster=NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before delete [] colorData\n");
    checkHeap(" Check Heap:");
#endif
  // Delete the color array data
    if(colorData) {
	delete [] colorData;
	colorData=NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before XDestroyImage\n");
    checkHeap(" Check Heap:");
    dumpHeap(" Dump Heap:");
#endif
  // Delete the image  (also deletes the imageData)
    if(image) {
      // Free the memory we allocated to insure it is done with the
      // same routines that allocated it.  Can be a problem on WIN32
      // if X frees it.
	if(image->data) free(image->data);
	image->data=NULL;
      // Then destroy the image
	XDestroyImage(image);
	image=NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before XFreeGC\n");
    checkHeap(" Check Heap:");
    dumpHeap(" Dump Heap:");
#endif
  // Free the gc
    if(gc) {
	XFreeGC(display,gc);
	gc=NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before XFreePixmap\n");
    checkHeap(" Check Heap:");
#endif
  // Free the pixmap
    if(pixmap) {
	XFreePixmap(display,pixmap);
	pixmap=(Pixmap)NULL;
    }
#if DEBUG_EXC
    print("~CImage: Before reset currentDrawingImage\n");
    checkHeap(" Check Heap:");
#endif
  // Reset the currentDrawingImage if it points to here
    if(currentDrawingImage == this) currentDrawingImage=NULL;
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CImage::storeSize(void)
{
    Byte *ptr=pGif->getPtr();
    Byte BlockSize,DataBlockSize;
    Byte ch;

  // Next 8- bytes are image data
    ptr+=8;

  // Check packed byte for local color table
    ch=NEXTBYTE;
    LocalColorTableFlag=ch & MASK7;
    SizeOfLocalColorTable=ch & MASK210;
    LocalColorTableEntries= 1 << (SizeOfLocalColorTable + 1);
    if(LocalColorTableFlag) ptr+=3*LocalColorTableEntries;

  // Skip LZW Minimum Code Size
    ptr++;

  // The rest is blocks
    introSize=ptr-pGif->getPtr();
    BlockSize=NEXTBYTE;
    ptr+=BlockSize;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }

    rawSize=(int)(ptr-pGif->getPtr());
}

int CImage::parse(void)
{
    Byte *ptr,ch;

  // Store the data
    int status=storeData();
    if(status != SGA_OK) return status;
    
    BitOffset=0;
    XC=0;
    YC=0;
    Pass=0;
    OutCount=0;
    
  // Store as text
    ptr=raw;
    if(storeText() != SGA_OK) {
	errMsg("CImage::parse: Cannot store text data",
	  stringType());
	return SGA_ERROR;
    }

  // Get parameters from the CGif
    nColors=pGif->getNColors();
    ScreenDepth=pGif->getScreenDepth();
    BytesPerPixel=pGif->getBytesPerPixel();
    BitMask=pGif->getBitMask();

  // Store as XImage
  // Read in values from the image descriptor
  // Parse the Left word
    ptr=raw;
    ch=NEXTBYTE;
    LeftOffset=ch + 0x100 * NEXTBYTE;
  // Parse the Top word
    ch=NEXTBYTE;
    TopOffset=ch + 0x100 * NEXTBYTE;
  // Parse the Width word
    ch=NEXTBYTE;
    Width=ch + 0x100 * NEXTBYTE;
  // Parse the Height word
    ch=NEXTBYTE;
    Height=ch + 0x100 * NEXTBYTE;
  // Parse the Packed byte
  // Bit 7:    MASK7   Local Color Table Flag
  // Bit 6:    MASK6   Interlace Flag
  // Bit 5:            Sort Flag (Ignored)
  // Bits 3-4:         Reserved
  // Bits 0-2: MASK210 Size of Local Color Table Entry
    ch=NEXTBYTE;
    LocalColorTableFlag=ch & MASK7;
    Interlace=((ch & MASK6) ? True : False);
    SizeOfLocalColorTable=ch & MASK210;
    LocalColorTableEntries= 1 << (SizeOfLocalColorTable + 1);

    if(LocalColorTableFlag) {
	errMsg("CImage::parse: Cannot currently handle local color table."
	  "  Using global table.");
      // Skip over it
	ptr+=3*LocalColorTableEntries;
    }

#if DEBUG_GIF
    print("CImage::parse: %dx%d+%d+%d, %d colors\n",
      Width,Height,LeftOffset,TopOffset,LocalColorTableEntries,
      (Interlace) ? "Interlaced" : "Non-Interlaced");
    if(LocalColorTableFlag) {
	print("  SizeOfLocalColorTable=%d\n",SizeOfLocalColorTable);
	print("  LocalColorTableEntries=%d\n",LocalColorTableEntries);
    }
#endif

  // Start reading the raster data. First we get the intial code size
  // and compute decompressor constant values, based on this code
  // size.
    CodeSize=NEXTBYTE;
    ClearCode=(1 << CodeSize);
    EOFCode=ClearCode + 1;
    FreeCode=FirstFree=ClearCode + 2;
    
  // The GIF spec has it that the code size is used to compute the
  // above values is the code size given in the file, but the code
  // size used in compression/decompression is the code size given in
  // the file plus one. (thus the ++).
    CodeSize++;
    InitCodeSize=CodeSize;
    MaxCode=(1 << CodeSize);
    ReadMask=MaxCode - 1;

#if DEBUG_IMAGE
    print("  CodeSize=%d InitCodeSize=%d ClearCode=%d EOFCode=%d "
      "MaxCode=%d FreeCode=%d\n",
      CodeSize-1,InitCodeSize,ClearCode,EOFCode,MaxCode,FreeCode);
#endif

  // Allocate memory for the raster data
    if(!(raster=new Byte[rawSize])) {
	errMsg("CImage::parse: Not enough memory for raster data");
	return SGA_ERROR;
    }

  // Allocate memory for the decompressor
    if(!(prefix=new int[4096])) {
	errMsg("CImage::parse: Not enough memory for decompressor prefix");
	return SGA_ERROR;
    }
    if(!(suffix=new int[4096])) {
	errMsg("CImage::parse: Not enough memory for decompressor suffix");
	return SGA_ERROR;
    }
    if(!(outCode=new int[1025])) {
	errMsg("CImage::parse: Not enough memory for decompressor outCode");
	return SGA_ERROR;
    }
    
  // Read the raster data block by block until the block terminator
  // (block with BlockSize=0).  Just transpose it from the GIF array
  // to the raster array, turning it from a series of blocks into one
  // long data stream, which makes life much easier for readCode().
    Byte *rptr=raster;
    int rasterBytes=0;
    Byte rch;
    do {
	ch=rch=NEXTBYTE;     // Number of bytes in block
	rasterBytes+=ch;
#if DEBUG_GIF > 1
	print("CImage::parse: rasterBytes=%d BlockSize=%d ch=%2x position=%d\n",
	  rasterBytes,rch,*ptr,ptr-RawGIF+1);
#endif
	while (ch--) *rptr++=NEXTBYTE;
	if((raster - rptr) > rawSize){
	    errMsg("CImage::parse: Trying to read past end of data");
	    return SGA_ERROR;
	}
    } while(rch);     // Continue until block terminator

#if DEBUG_GIF
    print("CImage::parse: decompressing...\n");
#endif

  // Allocate the X Image
  // (Used to be switch statement on ScreenDepth here)
    colorDataSize=Width*Height;
    imageDataSize=BytesPerPixel*colorDataSize;

  // Allocate the data for the XImage
    if(!imageData) {
	imageData=(Byte *)malloc(imageDataSize);
	if(!imageData) {
	    errMsg("CImage::parse: Not enough memory for XImage");
	    return SGA_ERROR;
	}
    }
  // Allocate the XImage.  The bitmap pad (next-to-last argument) may
  // be 8, 16, or 32.  The scanlines are padded out to an even number
  // of the bitmap_pad.  Use 8.  This should work with what we are
  // doing for depths of 8, 16, 24, 32, etc.  The bytes_per_line (last
  // argument) may be zero, indicating that X will calculate it for
  // you.  Otherwise it is probably width*bits_per_pixel/8.
    if(!image) {
	image=XCreateImage(display,pGif->getVisual(),
	  ScreenDepth,ZPixmap,0,
	  (char*)imageData,Width,Height,8,0);
	if(!image) {
	    errMsg("CImage::parse: Unable to create XImage");
	    return SGA_ERROR;
	}
	BytesPerScanline=image->bytes_per_line;
    }
  // We are always writing the data in MSBFirst
    image->byte_order=MSBFirst;
  // Allocate the color array data
    if(!colorData) {
	colorData=new Byte[colorDataSize];
	if(!colorData) {
	    errMsg("CImage::parse: Not enough memory for color array");
	    return SGA_ERROR;
	}
    }
#if DEBUG_IMAGE
    print("CImage::parse: [Image %d]\n",imageNumber);
    print("  width=%d\n",image->width);
    print("  height=%d\n",image->height);
    print("  xoffset=%d\n",image->xoffset);
    print("  byte_order=%s\n",
      image->byte_order==LSBFirst?"LSBFirst":"MSBFirst");
    print("  bitmap_unit=%d\n",image->bitmap_unit);
    print("  bitmap_bit_order=%s\n",
      image->bitmap_bit_order==LSBFirst?"LSBFirst":"MSBFirst");
    print("  bitmap_pad=%d\n",image->bitmap_pad);
    print("  depth=%d\n",image->depth);
    print("  bytes_per_line=%d\n",image->bytes_per_line);
    print("  bits_per_pixel=%d\n",image->bits_per_pixel);
#endif

  // Decompress the file, continuing until we see the GIF EOF code or
  // we use up all the bytes in the data */
    Code=readCode();
    while (Code != EOFCode && BitOffset/8 < rasterBytes) {
      // Clear code sets everything back to its initial value, then
      // reads the immediately subsequent code as uncompressed data.
	if(Code == ClearCode) {
	    CodeSize=InitCodeSize;
	    MaxCode=(1 << CodeSize);
	    ReadMask=MaxCode - 1;
	    FreeCode=FirstFree;
	    CurCode=OldCode=Code=readCode();
	    FinChar=CurCode & BitMask;
	    addToColorData((Byte)FinChar);
	} else {
	  // If not a clear code, then must be data: save same as CurCode and InCode
	    CurCode=InCode=Code;
	    
	  // If greater or equal to FreeCode, not in the hash table yet;
	  // repeat the last character decoded
	    if(CurCode >= FreeCode) {
		CurCode=OldCode;
		outCode[OutCount++]=FinChar;
	    }
	    
	  // Unless this code is raw data, pursue the chain pointed to by CurCode
	  // through the hash table to its end; each code in the chain puts its
	  // associated output code on the output queue.
	    while (CurCode > BitMask) {
		if(OutCount > 1024){
		    errMsg("CImage::parse: Corrupt GIF file (OutCount)");
		  // Free the memory we allocated to insure it is done with the
		  // same routines that allocated it.  Can be a problem on WIN32
		  // if X frees it.
		    if(image->data) free(image->data);
		    image->data=NULL;
		  // Then destroy the image
		    XDestroyImage(image);
		    image=NULL;
		    return SGA_ERROR;
		}
		outCode[OutCount++]=suffix[CurCode];
		CurCode=prefix[CurCode];
	    }
	    
	  // The last code in the chain is treated as raw data.
	    FinChar=CurCode & BitMask;
	    outCode[OutCount++]=FinChar;
	    
	  // Now we put the data out to the Output routine.  It's been
	  // stacked LIFO, so deal with it that way...
	    for(int i=OutCount - 1; i >= 0; i--)
	      addToColorData((Byte)outCode[i]);
	    OutCount=0;
	    
	  // Build the hash table on-the-fly. No table is stored in the file.
	    prefix[FreeCode]=OldCode;
	    suffix[FreeCode]=FinChar;
	    OldCode=InCode;
	    
	  // Point to the next slot in the table.  If we exceed the
	  // current MaxCode value, increment the code size unless
	  // it's already 12.  If it is, do nothing: the next code
	  // decompressed better be CLEAR
	    FreeCode++;
	    if(FreeCode >= MaxCode) {
		if(CodeSize < 12) {
		    CodeSize++;
		    MaxCode *= 2;
		    ReadMask=(1 << CodeSize) - 1;
		}
	    }
	}
	Code=readCode();
    }

  // Free the allocated data
    if(raster) {
	delete [] raster;
	raster=NULL;
    }
    if(prefix) {
	delete [] prefix;
	prefix=NULL;
    }
    if(suffix) {
	delete [] suffix;
	suffix=NULL;
    }
    if(outCode) {
	delete [] outCode;
	outCode=NULL;
    }

#if DEBUG_COLOR
	print("\nCImage::parse: colorDataSize=%d\n",colorDataSize);
	for(int i=0; i < colorDataSize; i++) {
	    Byte b=*(colorData+i);
	      print(" index=%3d %02x %02x %02x\n",
		b,pGif->getRed(b),pGif->getGreen(b),pGif->getBlue(b));
	}
#endif

  // Write the XImage data
    writeImageData();

    return SGA_OK;
}

int CImage::readCode(void)
{
    int RawCode,ByteOffset;

    ByteOffset=BitOffset/8;
    RawCode=raster[ByteOffset]+(0x100*raster[ByteOffset+1]);
    if(CodeSize >= 8)
      RawCode+=(0x10000*raster[ByteOffset+2]);
#if 0
    print("readCode: XC=%d YC=%d BitOffset=%d ByteOffset=%d Rawcode=%x",
      XC,YC,BitOffset,ByteOffset,RawCode);
#endif    
    RawCode>>=(BitOffset%8);
#if 0
    print("->%x ReadMask=%x RawCode&ReadMask=%x\n",
      RawCode,ReadMask,RawCode&ReadMask);
#endif    
    BitOffset+=CodeSize;
    return(RawCode&ReadMask);
}

// Writes a color to the color data corresponding to an index in the
// color table, keeping track of the rows and columns.  The color data
// are just the indices into the color table, without consideration of
// transparent color or background color.
void CImage::addToColorData(Byte index)
{
    int offset=YC*Width+XC;
	
#if DEBUG_COLOR
	print(" index=%3d %02x %02x %02x XC=%2d YC=%2d offset=%3d\n",
	  index,
	  pGif->getRed(index&(nColors-1)),pGif->getGreen(index&(nColors-1)),
	  pGif->getBlue(index&(nColors-1)),XC,YC,offset);
#endif
  // Check that we are in bounds and write to the color array
    if(offset < colorDataSize) {
	Byte *p=(Byte *)colorData+offset;
	*p=index&(nColors-1);
#if DEBUG_COLOR
	print("+index=%3d %02x %02x %02x XC=%2d YC=%2d offset=%3d\n",
	  *p,
	  pGif->getRed(*p&(nColors-1)),pGif->getGreen(*p&(nColors-1)),
	  pGif->getBlue(*p&(nColors-1)),XC,YC,offset);
#endif
    }
	
  // Update the X-coordinate, and if it overflows, update the Y-coordinate
    XC++;
    if(XC >= Width) {
      // If a non-interlaced picture, just increment YC to the next
      // scan line.  If it's interlaced, deal with the interlace as
      // described in the GIF spec.  Put the decoded scan line out to
      // the screen if we haven't gone past the bottom of it
	XC=0;
	if(!Interlace) {
	    YC++;
	} else {
	    switch (Pass) {
	    case 0:
		YC += 8;
		if(YC >= Height) {
		    Pass++;
		    YC=4;
		}
		break;
	    case 1:
		YC += 8;
		if(YC >= Height) {
		    Pass++;
		    YC=2;
		}
		break;
	    case 2:
		YC += 4;
		if(YC >= Height) {
		    Pass++;
		    YC=1;
		}
		break;
	    case 3:
		YC += 2;
		break;
	    default:
		break;
	    }
	}
    }
}

// Writes a color to the image data corresponding to an index in the
// color table, keeping track of the rows and columns.  The image data
// is pointed to by the XImage.  Changing the data changes the colors
// associated with the XImage.  The XImage needs to be copied to the
// pixmap for these changes to be visible.
void CImage::addToImageData(Byte index)
{
    Pixel clr=0;
    int offset=YC*BytesPerScanline+XC;
	
  //Check that we are in bounds and write to the XImage data
    if(offset < imageDataSize) {
	Byte *p=(Byte *)imageData+offset;

      //Check for transparent color
	if(getTransparentColorFlag() && (index&(nColors-1)) ==
	  getTransparentIndex()) {
	    switch(pGif->getBgMode()) {
	    case BACKGROUND_GIF:
		clr=pGif->getBackgroundColor();
		break;
	    case BACKGROUND_WHITE:
		clr=WhitePixel(display,screen);
		break;
	    case BACKGROUND_BLACK:
		clr=BlackPixel(display,screen);
		break;
	    }
	} else {
	    clr=pixels[index&(nColors-1)];
	}
	
      // Write the appropriate part of the pixel in MSBFirst order
	switch(BytesPerPixel) {
	case 4:
	    *p++=(Byte)((clr & 0xff000000) >> 24);
	  /* Fall through */
	case 3:
	    *p++=(Byte)((clr & 0x00ff0000) >> 16);
	  /* Fall through */
	case 2:
	    *p++=(Byte)((clr & 0x0000ff00) >> 8);
	  /* Fall through */
	case 1:
	    *p=(Byte)((clr & 0x000000ff));
	    break;
	}
    }

  //Update the X-coordinate, and if it overflows, update the Y-coordinate
    XC=XC+BytesPerPixel;
    if(XC+BytesPerPixel > BytesPerScanline) {
      // Do interlace in the colorData, not here
	XC=0;
	YC++;
    }
}

// Writes the color array data to the image data, taking into account
// the transparent color and the background color.  The current XImage
// will be updated, but the pixmap will not be.
void CImage::writeImageData(void) {
    Byte *iptr=colorData;
    Byte *eof=colorData+colorDataSize;

#if DEBUG_COLOR
    print("CImage::writeImageData: colorDataSize=%d imageDataSize=%d\n",
      colorDataSize,imageDataSize);
#endif

  // Get the current pixel array
    nColors=pGif->getNColors();
    pixels=pGif->getPixels();

  // Just copy the color indices from the colorArray to the imagedata
    XC=0;
    YC=0;
    Pass=0;
    while(iptr < eof) {
#if DEBUG_COLOR
	print(" index=%3d %02x %02x %02x\n",
	  *iptr,
	  pGif->getRed(*iptr&(nColors-1)),pGif->getGreen(*iptr&(nColors-1)),
	  pGif->getBlue(*iptr&(nColors-1)));
#endif
	addToImageData(*iptr);
	iptr++;
    }
}

int CImage::write(FILE *fp)
{

    Byte b=0;

  // First let the control block write itself
    if(control) control->write(fp);

  // Write the image separator
    fputc(IMAGESEP,fp);

  // Write the sizes
    writeWord(LeftOffset,fp);
    writeWord(TopOffset,fp);
    writeWord(Width,fp);
    writeWord(Height,fp);

  // Write the Packed byte
  //   Bit 7:    MASK7   Local Color Table Flag (Ignored)
  //   Bit 6:    MASK6   Interlace Flag
  //   Bit 5:            Sort Flag (Ignored)
  //   Bits 3-4:         Reserved (Ignored)
  //   Bits 0-2: MASK210 Size of Local Color Table Entry (Ignored)
    if(Interlace) b|=MASK6;
    fputc(b,fp);

  // Write the initial code size
    int bitsPerPixel=pGif->getBitsPerPixel();
    if(bitsPerPixel <= 1)
      InitCodeSize=2;
    else
      InitCodeSize=bitsPerPixel;
    fputc(InitCodeSize,fp);

  // Compress the data and write it
    sgaCompress(fp,InitCodeSize,colorData,Width,Height,Interlace?1:0);

  // Write the final block terminator
    fputc(BLOCKTERM,fp);

  //    if(rawSize) fwrite(raw+9,rawSize-9,1,fp);

    return SGA_OK;
}

int CImage::displayData(void)
{
    char line[81];
    
  // Update parameters
    nColors=pGif->getNColors();
    pixels=pGif->getPixels();

    if(pGif->getDataMode() == TEXT_MODE) {
      // TEXT_MODE
	changeDisplayWindow(TEXT_MODE);
	
      // Set the option area
	displayOptions();

      // Put the Control text in the drawing area
	if(control) control->displayData();
	
      // Put the Image text in the drawing area
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	  "Image Block\n\n");
	sprintf(line,"Width:                     %d\n",Width);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"Height:                    %d\n",Height);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"LeftOffset:                %d\n",LeftOffset);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"TopOffset:                 %d\n",TopOffset);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	
	
	sprintf(line,"Local Color Table Flag:    %s\n",
	  LocalColorTableFlag?"True":"False");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"Size Of Local Color Table: %d\n",
	  SizeOfLocalColorTable);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"Local Color Table Entries: %d\n",
	  LocalColorTableEntries);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	sprintf(line,"Interlace:                 %s\n\n",
	  Interlace?"True":"False");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
	
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	  "Data for Image Block\n\n");
	if(text) {
	    XmTextInsert(textArea,XmTextGetLastPosition(textArea),text);
	} else {
	    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	      "No information\n");
	}
	XmTextSetCursorPosition(textArea,0);
	XtManageChild(textArea);
    } else {
      // GRAPHIC_MODE
	changeDisplayWindow(GRAPHIC_MODE);
	currentDrawingImage=this;

      // Set the option area
	displayOptions();

      // Make the pixmap
	int status=makePixmap();
	if(status != SGA_OK) return status;

      // Run the exposeCB to copy the pixmap
	exposeCB();

      // Remanage the drawingSW to get sizes right
	XtManageChild(drawingSW);
    }
      
    return SGA_OK;
}

// Makes a pixmap with size the larger of the image or the drawing
// area and puts the current XImage on it.
int CImage::makePixmap(void)
{
      // Get the drawingSW's dimenions
	Dimension sWidth,sHeight;
	nargs=0;
	XtSetArg(args[nargs],XmNwidth,&sWidth); nargs++;
	XtSetArg(args[nargs],XmNheight,&sHeight); nargs++;
	XtGetValues(drawingSW,args,nargs);

      // Get the image's full dimension
	Dimension iWidth=pGif->getLogicalScreenWidth();
	Dimension iHeight=pGif->getLogicalScreenHeight();

      // Resize the drawing area
	pWidth=MAX(iWidth,sWidth);
	pHeight=MAX(iHeight,sHeight);
	nargs=0;
	XtSetArg(args[nargs],XmNwidth,pWidth); nargs++;
	XtSetArg(args[nargs],XmNheight,pHeight); nargs++;
	XtSetValues(drawingArea,args,nargs);
	XtManageChild(drawingArea);

      // Free any existing pixmap and make a new one
	if(pixmap) {
	    XFreePixmap(display,pixmap);
	    pixmap=(Pixmap)NULL;
	}
	pixmap=XCreatePixmap(display,RootWindow(display,screen),pWidth,pHeight,
	  pGif->getScreenDepth());
	if(!pixmap) {
	    errMsg("CImage::makePixmap: Cannot create pixmap");
	    return SGA_ERROR;
	}

      // Fill pixmap with the background color
	switch(pGif->getBgMode()) {
	case BACKGROUND_GIF:
	    XSetForeground(display,gc,pGif->getBackgroundColor());
	    break;
	case BACKGROUND_WHITE:
	    XSetForeground(display,gc,WhitePixel(display,screen));
	    break;
	case BACKGROUND_BLACK:
	    XSetForeground(display,gc,BlackPixel(display,screen));
	    break;
	}
	XFillRectangle(display,pixmap,gc,0,0,pWidth,pHeight);	

      // Copy the image to the pixmap
	Position iX=0;
	Position iY=0;

	if(pWidth > iWidth) iX=(pWidth-iWidth)/2;
	if(pHeight > iHeight) iY=(pHeight-iHeight)/2;
	iX+=LeftOffset;
	iY+=TopOffset;

#if DEBUG_DRAW
	print("CImage::makePixmap:\n  sWidth=%d sHeight=%d\n",sWidth,sHeight);
	print("  iWidth=%d iHeight=%d\n",iWidth,iHeight);
	print("  pWidth=%d pHeight=%d\n",pWidth,pHeight);
	print("  iX=%d iY=%d\n",iX,iY);
	print("  pixmap=%x currentDrawingImage=%x this=%x\n",
	  pixmap,currentDrawingImage,this);
#endif

	XPutImage(display,pixmap,gc,image,0,0,iX,iY,Width,Height);	

	return SGA_OK;
}

// Handles drawing area expose events
void CImage::exposeCB(void)
{
#if DEBUG_DRAW
    print("CImage::exposeCB: pixmap=%x\n",pixmap);
#endif
  // Just copy the pixmap to the drawing area
    if(pixmap) {
	XCopyArea(display,pixmap,XtWindow(drawingArea),gc,0,0,
	  pWidth,pHeight,0,0);
    }
}

// Handles drawing area resize events
void CImage::resizeCB(void)
{
  // Redo the data display to get resized pixmap, drawingArea
    displayData();
}

int CImage::displayOptions(void)
{
  // Update the options
    updateOptionsArea();

    XtManageChild(imageOptionArea);

  // Unmanage any existing option area
    if(globalOptionArea && XtIsManaged(globalOptionArea))
      XtUnmanageChild(globalOptionArea);
    if(defaultOptionArea && XtIsManaged(defaultOptionArea))
      XtUnmanageChild(defaultOptionArea);
    if(commentOptionArea && XtIsManaged(commentOptionArea))
      XtUnmanageChild(commentOptionArea);

    return SGA_OK;
}

void CImage::updateOptionsArea(void) const
{
    char string[81];

#if DEBUG_TOGGLE
	Dimension marginHeight,marginTop,marginBottom,height;
	nargs=0;
	XtSetArg(args[nargs],XmNmarginHeight,&marginHeight); nargs++;
	XtSetArg(args[nargs],XmNmarginTop,&marginTop); nargs++;
	XtSetArg(args[nargs],XmNmarginBottom,&marginBottom); nargs++;
	XtSetArg(args[nargs],XmNheight,&height); nargs++;
	XtGetValues(interlaceBtn,args,nargs);
	print("CImage::updateOptionsArea:"
	  " marginHeight=%d marginTop=%d marginBottom=%d height=%d\n",
	  marginHeight,marginTop,marginBottom,height);
#endif

#ifdef WIN32
      // Set the margins in WIN32  (shouldn't be necessary)
	nargs=0;
	XtSetArg(args[nargs],XmNmarginTop,0); nargs++;
	XtSetArg(args[nargs],XmNmarginBottom,0); nargs++;
	XtSetValues(applyToAllImagesBtn,args,nargs);
	XtSetValues(transparentColorBtn,args,nargs);
	XtSetValues(interlaceBtn,args,nargs);
	XtSetValues(localColorTableBtn,args,nargs);
#endif

  // Toggle buttons
    XmToggleButtonSetState(transparentColorBtn,
      getTransparentColorFlag()?True:False,False);
    XmToggleButtonSetState(interlaceBtn,
      Interlace?True:False,False);
    XmToggleButtonSetState(localColorTableBtn,
      LocalColorTableFlag?True:False,False);

  // Set the userData
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetValues(transparentColorBtn,args,nargs);
    XtSetValues(interlaceBtn,args,nargs);
    XtSetValues(localColorTableBtn,args,nargs);
    XtSetValues(disposalMethodMenu,args,nargs);

  // Text entries
    sprintf(string,"%d",control->TransparentIndex);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(transparentIndexEntry,args,nargs);

    sprintf(string,"%d",control->DelayTime);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(delayTimeEntry,args,nargs);

    sprintf(string,"%d",LeftOffset);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(leftOffsetEntry,args,nargs);

    sprintf(string,"%d",TopOffset);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(topOffsetEntry,args,nargs);

  // Disposal method
    int method=control->DisposalMethod;
    if(method >= 0 || method < 4) {
	disposalMethodSet(control->DisposalMethod);
    } else {
	disposalMethodSet(0);
    }
}

void CImage::setTransparentColorFlag(Boolean flag)
{
    control->TransparentColorFlag=flag?1:0;
    updateOptionsArea();
    writeImageData();
}

void CImage::setInterlace(Boolean flag)
{
    Interlace=flag;
    updateOptionsArea();
}

void CImage::setLocalColorTableFlag(Boolean /* flag */)
{
    errMsg("Setting the local color table is not implemented");
    updateOptionsArea();
}

void CImage::setTransparentIndex(char *value)
{
  // Has to be one byte, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xff) val=0xff;
    control->TransparentIndex=(Byte)val;
    updateOptionsArea();
    writeImageData();
}

void CImage::setDelayTime(char *value)
{
  // Has to be two bytes, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xffff) val=0xffff;
    control->DelayTime=val;
    updateOptionsArea();
}

void CImage::setLeftOffset(char *value)
{
  // Has to be two bytes, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xffff) val=0xffff;
    LeftOffset=val;
    updateOptionsArea();
}

void CImage::setTopOffset(char *value)
{
  // Has to be two bytes, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xffff) val=0xffff;
    TopOffset=val;
    updateOptionsArea();
}

void CImage::setDisposalMethod(int method)
{
    if(method >= 0 || method < 8) {
	control->DisposalMethod=method;
    } else {
	control->DisposalMethod=0;
    }
    updateOptionsArea();
}

Boolean CImage::isUsedColor(Byte index)
{
    for(int i=0; i < colorDataSize; i++) {
	if(index == colorData[i]) return True;
    }

    return False;
}

// Change the color indices according to a translation table
void CImage::translateColors(Byte *table)
{
  // Change the color data
    for(int i=0; i < colorDataSize; i++) {
	colorData[i]=table[colorData[i]&0xff];
    }

  // Change the transparent color index
    control->TransparentIndex=table[control->TransparentIndex&&0xff];
}


char *CImage::getDisposalMethodString(int n)
{
    if(n >= 0 || n < 8) {
	return stringDisposalMethod[n];
    } else {
	return stringDisposalMethod[7];
    }
}

/**************************************************************************/
/**************************** CPlaintext **********************************/
/**************************************************************************/

CPlaintext::CPlaintext(CGif *parent) :
    CBlock(parent,TYPE_PLAINTEXT)
{
}

CPlaintext::~CPlaintext(void)
{
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CPlaintext::storeSize(void)
{
    Byte *ptr=pGif->getPtr();
    Byte BlockSize,DataBlockSize;

    BlockSize=NEXTBYTE;     /* Should be 12 */
    introSize=BlockSize+1;
    ptr+=BlockSize;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }

    rawSize=(int)(ptr-pGif->getPtr());
}

int CPlaintext::parse(void)
{
  // Store the data
    int status=storeData();
    if(status != SGA_OK) return status;

#if 0
      // Ignore
	BlockSize=NEXTBYTE;     /* Should be 12 */
	ptr+=BlockSize;
	while(DataBlockSize=NEXTBYTE) {
	    ptr+=DataBlockSize;
	}
#endif	

  // Store as text
    if(storeText() != SGA_OK) {
	errMsg("CPlaintext::parse: Cannot store text data",
	  stringType());
	return SGA_ERROR;
    }
    return SGA_OK;
}

int CPlaintext::write(FILE *fp)
{
  // Check if there is data
    if(!raw || !rawSize) return SGA_ERROR;

  // Write extension introducer and extension label
    fputc(EXTENSION,fp);
    fputc(PLAINTEXT,fp);

  // Just copy the raw data
    fwrite(raw,rawSize,1,fp);

    return SGA_OK;
}

int CPlaintext::displayData(void)
{
  // Change the scrolled window that is managed
    changeDisplayWindow(TEXT_MODE);

  // Set the option area
    displayOptions();

  // Put the text in the drawing area
    if(text) {
	XmTextSetString(textArea,"Data for Plaintext Block\n\n");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),text);
    } else {
	XmTextSetString(textArea,"No information\n");
    }
    XmTextSetCursorPosition(textArea,0);
    XtManageChild(textArea);
      
    return SGA_OK;
}

int CPlaintext::displayOptions(void)
{
    XtManageChild(defaultOptionArea);

  // Unmanage any existing option area
    if(globalOptionArea && XtIsManaged(globalOptionArea))
      XtUnmanageChild(globalOptionArea);
    if(imageOptionArea && XtIsManaged(imageOptionArea))
      XtUnmanageChild(imageOptionArea);
    if(commentOptionArea && XtIsManaged(commentOptionArea))
      XtUnmanageChild(commentOptionArea);

    return SGA_OK;
}

/**************************************************************************/
/**************************** CApplication ********************************/
/**************************************************************************/

CApplication::CApplication(CGif *parent) :
    CBlock(parent,TYPE_APPLICATION)
{
}

CApplication::~CApplication(void)
{
    *ApplicationIdentifier='\0';
    *ApplicationAuthenticationCode='\0';
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CApplication::storeSize(void)
{
    Byte *ptr=pGif->getPtr();
    Byte BlockSize,DataBlockSize;

    BlockSize=NEXTBYTE;
    introSize=BlockSize+1;
    ptr+=BlockSize;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }

    rawSize=(int)(ptr-pGif->getPtr());
}

int CApplication::parse(void)
{
    Byte BlockSize,DataBlockSize;
    Byte *ptr;

  // Store the data
    int status=storeData();
    if(status != SGA_OK) return status;

  // Parse the intro block
    ptr=raw;
    BlockSize=NEXTBYTE;     // Should be 0x0B=11
    strncpy(ApplicationIdentifier,(char *)ptr,8);
    ApplicationIdentifier[8]='\0';
    strncpy(ApplicationAuthenticationCode,(char *)(ptr+8),3);
    ApplicationAuthenticationCode[3]='\0';
    
  // Skip over the rest
    ptr+=BlockSize;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }

  // Store as text
    if(storeText() != SGA_OK) {
	errMsg("CApplication::parse: Cannot store text data",
	  stringType());
	return SGA_ERROR;
    }

    return SGA_OK;
}

int CApplication::write(FILE *fp)
{
  // Check if there is data
    if(!raw || !rawSize) return SGA_ERROR;

  // Write extension introducer and extension label
    fputc(EXTENSION,fp);
    fputc(APPLICATION,fp);

  // Just copy the raw data
    fwrite(raw,rawSize,1,fp);

    return SGA_OK;
}

int CApplication::displayData(void)
{
  // Change the scrolled window that is managed
    changeDisplayWindow(TEXT_MODE);

  // Set the option area
    displayOptions();

  // Put the intro data in the textArea
    XmTextSetString(textArea,"Application Identifier:          ");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
      ApplicationIdentifier);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
      "\nApplication Authentication Code: ");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
      ApplicationAuthenticationCode);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
      "\n\n");
    
  // Put the text in the drawing area
    if(text) {
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	  "Data for Application Block\n\n");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),text);
    } else {
	XmTextSetString(textArea,"No information\n");
    }
    XmTextSetCursorPosition(textArea,0);
    XtManageChild(textArea);
      
    return SGA_OK;
}

int CApplication::displayOptions(void)
{
    XtManageChild(defaultOptionArea);

  // Unmanage any existing option area
    if(globalOptionArea && XtIsManaged(globalOptionArea))
      XtUnmanageChild(globalOptionArea);
    if(imageOptionArea && XtIsManaged(imageOptionArea))
      XtUnmanageChild(imageOptionArea);
    if(commentOptionArea && XtIsManaged(commentOptionArea))
      XtUnmanageChild(commentOptionArea);

    return SGA_OK;
}

/**************************************************************************/
/**************************** CComment ************************************/
/**************************************************************************/

CComment::CComment(CGif *parent) :
    CBlock(parent,TYPE_COMMENT),
    comment(NULL)
{
}

CComment::~CComment(void)
{
  // Free the comment
    if(comment) {
	delete [] comment;
	comment=NULL;
    }
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CComment::storeSize(void)
{
    Byte *ptr=pGif->getPtr();
    Byte DataBlockSize;

  /* There is no block before the data here */
    introSize=0;
    while((DataBlockSize=NEXTBYTE)) {
	ptr+=DataBlockSize;
    }
    rawSize=(int)(ptr-pGif->getPtr());
}

int CComment::parse(void)
{
  // Store the data
    int status=storeData();
    if(status != SGA_OK) return status;

  // Store as text
    if(storeText() != SGA_OK) {
	errMsg("CComment::parse: Cannot store text data",
	  stringType());
	return SGA_ERROR;
    }
    return SGA_OK;
}

int CComment::write(FILE *fp)
{
    int i,j,size;

  // Check if there is a comment
    if(!comment || !*comment) return SGA_ERROR;

  // Note that the comment array may be longer than necessary
    int stringSize=strlen(comment);

  // Write extension introducer and extension label
    fputc(EXTENSION,fp);
    fputc(COMMENT,fp);

  // Comments appear to have CR-LF separators so convert ours, which only has LF
  // First determine the size
    size=0;
    for(i=0; i < stringSize; i++) {
	size++;
	if(comment[i] == 0x0a) size++;
    }

  // Allocate a new, extended array
    char *data=new char[size];
    if(!data) {
	errMsg("CComment::write: Cannot allocate space for extended comment");
	return SGA_ERROR;
    }

  // Add the CRs
    for(i=0,j=0; i < stringSize; i++) {
	if(comment[i] == 0x0a) data[j++]=0x0d;
	data[j++]=comment[i];
    }

  // Write the blocks
#if DEBUG_COMMENT
	char *eod=data+size;
	print("CComment::write commentSize=%d stringSize=%d str\n",
	  commentSize,stringSize);
#endif
    char *pData=data;
    while(size > 0) {
	int blockSize;
	if(size > 255) blockSize=255;
	else blockSize=size;
      // Write the block size
	fputc(blockSize,fp);
      // Write the block
	fwrite(pData,blockSize,1,fp);
#if DEBUG_COMMENT
	print("  pData=%x->%x eod=%d size=%d blockSize=%d\n",
	  pData,pData+blockSize,eod,size,blockSize);
	for(int k=0; k < blockSize; k++) {
	    char ch=*(pData+k);
	    if(ch > 31 && ch < 128) {
		print("%c",ch);
	    } else if(ch == 0x0d) {
		print("\\r");
	    } else if(ch == 0x0a) {
		print("\\l");
		print("\n");
	    } else {
		print("(%d)",(int)ch);
	    }
	}
	print("|\n");
#endif
      // Recalculate the size;
	pData+=blockSize;
	size-=blockSize;
    }
    
  // Free the extended array
    delete [] data;

  // Write the block terminator;
    fputc(BLOCKTERM,fp);

    return SGA_OK;
}

int CComment::displayData(void)
{
  // Change the scrolled window that is managed (TEXT_MODE both cases)
    changeDisplayWindow(TEXT_MODE);

  // Set the option area
    displayOptions();

  // Display text or comment
    if(pGif->getDataMode() == TEXT_MODE) {
      // TEXT_MODE
	XmTextSetString(textArea,"");
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	  "Data for Comment Block\n\n");
	if(text) {
	    XmTextInsert(textArea,XmTextGetLastPosition(textArea),text);
	} else {
	    XmTextInsert(textArea,XmTextGetLastPosition(textArea),
	      "No information\n");
	}
    } else {
      // GRAPHIC_MODE
	if(comment) {
	    XmTextSetString(textArea,comment);
	} else {
	    XmTextSetString(textArea,"<No comment>\n");
	}
    }
    XmTextSetCursorPosition(textArea,0);
    XtManageChild(textArea);
    
    return SGA_OK;
}

int CComment::storeText(void)
{
    Byte *ptr;
    Byte ch;
    char *tptr;
    Byte DataBlockSize;
    const int dataSize=rawSize-introSize;
    int nBlock,nTotal,lines;
    int status;

  // Store the standard text
    status=CBlock::storeText();
    if(status != SGA_OK) return status;

  // Store the comment
    if(dataSize <= 0) return SGA_ERROR;

  // Set the following to add line breaks
    int lineBreaks=0;
    
  // Calculate the text size needed. (Allow an extra line.)
    if(lineBreaks) {
	lines=dataSize/LINESIZE+2;
	commentSize=dataSize+lines+1;     // For '\n' for each line
    } else {
	commentSize=dataSize+1;
    }

  // Allocate the space
    if(comment) delete [] comment;
    comment=new char[commentSize];
    if(!comment) {
	errMsg("CComment::storeText [%s]: Cannot allocate space for comment",
	  stringType());
	return SGA_ERROR;
    }

  // Get the bytes from the data blocks
    ptr=raw+introSize;       // Start of data blocks
    tptr=comment;               // Start of comment
    nTotal=0;
    while((DataBlockSize=NEXTBYTE)) {
	nBlock=0;
	while(nBlock < DataBlockSize) {
	    ch=NEXTBYTE;
	    if(lineBreaks && nTotal && nTotal%LINESIZE == 0) {
		*tptr++='\n';
		nTotal=0;
	    }
	    nBlock++; nTotal++;
	    if(ch > 31 && ch < 128) {
	      // Is printable ASCII
		*tptr++=ch;
	    } else if(ch == 0x0D) {    // CR
	      // *tptr++='\r';
		nTotal=0;
	    } else if(ch == 0x0A) {    // CF
		*tptr++='\n';
		nTotal=0;
	    } else {
		*tptr++='.';
	    }
	}
	*tptr='\0';
    }
#if DEBUG_TEXT
    printf(comment);      // May be too big for print
#endif

    return SGA_OK;
}

int CComment::displayOptions(void)
{

  // Put the comment in the text area
    XmTextSetString(commentText,comment);
    XmTextSetCursorPosition(commentText,0);

  // Set the user data
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetValues(commentApplyBtn,args,nargs);

    XtManageChild(commentOptionArea);

  // Unmanage any existing option area
    if(defaultOptionArea && XtIsManaged(defaultOptionArea))
      XtUnmanageChild(defaultOptionArea);
    if(globalOptionArea && XtIsManaged(globalOptionArea))
      XtUnmanageChild(globalOptionArea);
    if(imageOptionArea && XtIsManaged(imageOptionArea))
      XtUnmanageChild(imageOptionArea);

    return SGA_OK;
}

int CComment::setComment(char *value)
{
    if(!value) return SGA_ERROR;

  // Allocate the space
    int len=strlen(value);
    if(comment) delete [] comment;
    comment=new char[len+1];
    if(!comment) {
	errMsg("CComment::setComment: Cannot allocate space for comment");
	return SGA_ERROR;
    }
    strcpy(comment,value);
    commentSize=len+1;

    return SGA_OK;
}

/**************************************************************************/
/**************************** CGlobal *************************************/
/**************************************************************************/

CGlobal::CGlobal(CGif *parent) :
    CBlock(parent,TYPE_GLOBAL)
{
}

CGlobal::~CGlobal(void)
{
}

// Stores the size of the raw data in the file for this block,
// including the block size bytes in rawData
void CGlobal::storeSize(void)
{
    rawSize=0;
}

int CGlobal::parse(void)
{
    return SGA_OK;
}

int CGlobal::write(FILE *fp)
{
    Byte b=0;

  // Write the signature
  // KE: Fix this later to check if it may we written as GIF87a
    fwrite("GIF89a",6,1,fp);

  // Write the Logical Screen Descriptor

  // Write the Width and height
    writeWord(pGif->LogicalScreenWidth,fp);
    writeWord(pGif->LogicalScreenHeight,fp);

  // Write the packed byte
  //   Bits 0-2: MASK210 Size of the Global Color Table
  //   Bit 3:            Color Table Sort Flag (Ignored)
  //   Bits 4-6:         Color Resolution
  //   Bit 7:    MASK7   Global Color Table Flag
    if(pGif->GlobalColorTableFlag) b|=MASK7;
  // Set the ColorResolution to the ScreenDepth minus 1
    int resolution=pGif->ScreenDepth-1;
    b|=((resolution<<4)&MASK654);
    b|=(pGif->SizeOfGlobalColorTable&MASK210);
    fputc(b,fp);

  // Write the BackgroundColorIndexColor
    fputc((Byte)pGif->BackgroundColorIndex,fp);

  // Write the PixelAspectRatio (Use 0)
    Byte aspect=0;
    fputc(aspect,fp);

  // Write the global color map
    for(int i=0; i < pGif->nColors; i++) {
	fputc(pGif->Red[i],fp);
	fputc(pGif->Green[i],fp);
	fputc(pGif->Blue[i],fp);
    }

    return SGA_OK;
}

int CGlobal::displayData(void)
{
    char line[PATH_MAX+10];  // 6 extra characters should be enough

#if DEBUG_DRAW
    {
	Widget workWindow=NULL,vsb=NULL,hsb=NULL;
	Dimension sWidth,sHeight;
	
	print("CGlobal::displayData:"
	  " textSW=%x drawingSW=%x textArea=%x drawingArea=%x\n",
	  textSW,drawingSW,textArea,drawingArea);

	nargs=0;
	XtSetArg(args[nargs],XmNwidth,&sWidth); nargs++;
	XtSetArg(args[nargs],XmNheight,&sHeight); nargs++;
	XtSetArg(args[nargs],XmNhorizontalScrollBar,&hsb); nargs++;
	XtSetArg(args[nargs],XmNverticalScrollBar,&vsb); nargs++;
	XtSetArg(args[nargs],XmNworkWindow,&workWindow); nargs++;
	XtGetValues(textSW,args,nargs);
	print("textSW:  workWindow=%x sWidth=%d sHeight=%d hsb=%x vsb=%x\n",
	  workWindow,sWidth,sHeight,hsb,vsb);
	print("  IsManaged(textgSW)=%s\n",XtIsManaged(textSW)?"Yes":"No");
	if(hsb) print("  IsManaged(hsb)=%s\n",XtIsManaged(hsb)?"Yes":"No");
	if(vsb) print("  IsManaged(vsb)=%s\n",XtIsManaged(vsb)?"Yes":"No");

	nargs=0;
	XtSetArg(args[nargs],XmNwidth,&sWidth); nargs++;
	XtSetArg(args[nargs],XmNheight,&sHeight); nargs++;
	XtSetArg(args[nargs],XmNhorizontalScrollBar,&hsb); nargs++;
	XtSetArg(args[nargs],XmNverticalScrollBar,&vsb); nargs++;
	XtSetArg(args[nargs],XmNworkWindow,&workWindow); nargs++;
	XtGetValues(drawingSW,args,nargs);
	print("drawingSW:  workWindow=%x sWidth=%d sHeight=%d hsb=%x vsb=%x\n",
	  workWindow,sWidth,sHeight,hsb,vsb);
	print("  IsManaged(drawingSW)=%s\n",XtIsManaged(drawingSW)?"Yes":"No");
	if(hsb) print("  IsManaged(hsb)=%s\n",XtIsManaged(hsb)?"Yes":"No");
	if(vsb) print("  IsManaged(vsb)=%s\n",XtIsManaged(vsb)?"Yes":"No");

	if(textArea) {
	    Dimension tWidth,tHeight;

	    nargs=0;
	    XtSetArg(args[nargs],XmNwidth,&tWidth); nargs++;
	    XtSetArg(args[nargs],XmNheight,&tHeight); nargs++;
	    XtGetValues(textArea,args,nargs);
	    print("IsManaged(textArea)=%s\n",
	      XtIsManaged(textArea)?"Yes":"No");
	    print("  tWidth=%d tHeight=%d\n",tWidth,tHeight);
	}
	if(drawingArea) {
	    Dimension dWidth,dHeight;

	    nargs=0;
	    XtSetArg(args[nargs],XmNwidth,&dWidth); nargs++;
	    XtSetArg(args[nargs],XmNheight,&dHeight); nargs++;
	    XtGetValues(drawingArea,args,nargs);
	    print("IsManaged(drawingArea)=%s\n",
	      XtIsManaged(drawingArea)?"Yes":"No");
	    print("  dWidth=%d dHeight=%d\n",dWidth,dHeight);
	}
    }
#endif

  // Change the scrolled window that is managed
    changeDisplayWindow(TEXT_MODE);

  // Set the option area
    displayOptions();

  // Put the intro data in the textArea
    XmTextSetString(textArea,"");
    sprintf(line,"File: %s\n",pGif->fileName);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"FileSize:                     %d\n",pGif->fileSize);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Type:                         %s\n\n",pGif->GifType);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    sprintf(line,"Logical Screen Width:         %d\n",
      pGif->LogicalScreenWidth);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Logical Screen Height:        %d\n\n",
      pGif->LogicalScreenHeight);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    sprintf(line,"Global Color Table Flag:      %s\n",
      pGif->GlobalColorTableFlag?"True":"False");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Global Color Table Sort Flag: %s\n",
      pGif->GlobalColorTableSortFlag?"True":"False");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Color Resolution:             %d\n",
      pGif->ColorResolution);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Color Table Entries:          %d\n",
      pGif->ColorTableEntries);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Bits Per Pixel:               %d [%d colors]\n",
      pGif->BitsPerPixel,pGif->ColorTableEntries);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Background Color Index:       %d\n\n",
      pGif->BackgroundColorIndex);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    sprintf(line,"Pixel Aspect Ratio:           %d\n\n",
      pGif->PixelAspectRatio);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    sprintf(line,"Screen Depth:                 %d [%d colors]\n",
      pGif->ScreenDepth,1<<pGif->ScreenDepth);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"X Bits Per Pixel:             %d\n",pGif->XBitsPerPixel);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"X Bitmap Bit Order:           %s\n",
      pGif->XBitmapBitOrder==LSBFirst?"LSBFirst":"MSBFirst");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"X Bitmap Pad:                 %d\n",pGif->XBitmapPad);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"X Bitmap Unit:                %d\n",pGif->XBitmapUnit);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"X Byte Order:                 %s\n\n",
      pGif->XByteOrder==LSBFirst?"LSBFirst":"MSBFirst");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    sprintf(line,"Client Byte Order:            %s\n",
      pGif->ClientByteOrder==LSBFirst?"LSBFirst":"MSBFirst");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    sprintf(line,"Bytes Offset Per Pixel:       %d\n\n",pGif->BytesPerPixel);
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);

    XmTextSetCursorPosition(textArea,0);
    XtManageChild(textArea);
      
    sprintf(line,"Global Color Table:\n     R  G  B\n");
    XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    for(int i=0; i < pGif->nColors; i++) {
	sprintf(line,"%3d %02x %02x %02x\n",i,
	  pGif->Red[i],pGif->Green[i],pGif->Blue[i]);
	XmTextInsert(textArea,XmTextGetLastPosition(textArea),line);
    }

    return SGA_OK;
}

int CGlobal::storeText(void)
{
    return SGA_OK;
}

int CGlobal::displayOptions(void)
{
  // Update the options
    updateOptionsArea();

    XtManageChild(globalOptionArea);

  // Unmanage any existing option area
    if(defaultOptionArea && XtIsManaged(defaultOptionArea))
      XtUnmanageChild(defaultOptionArea);
    if(imageOptionArea && XtIsManaged(imageOptionArea))
      XtUnmanageChild(imageOptionArea);
    if(commentOptionArea && XtIsManaged(commentOptionArea))
      XtUnmanageChild(commentOptionArea);

    return SGA_OK;
}

void CGlobal::updateOptionsArea(void) const
{
    char string[81];

  // Text entries
    sprintf(string,"%d",pGif->BackgroundColorIndex);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(bgColorIndexEntry,args,nargs);

    sprintf(string,"%d",pGif->LogicalScreenWidth);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(logicalScreenWidthEntry,args,nargs);

    sprintf(string,"%d",pGif->LogicalScreenHeight);
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;
    XtSetArg(args[nargs],XmNcursorPosition,0); nargs++;
    XtSetArg(args[nargs],XmNvalue,string); nargs++;
    XtSetValues(logicalScreenHeightEntry,args,nargs);
}

void CGlobal::setBackgroundColorIndex(char *value)
{
  // Has to be one byte, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xff) val=0xff;
    pGif->BackgroundColorIndex=val;
    pGif->BackgroundColor=pGif->pixels[val];
    updateOptionsArea();
    displayData();

  // Redo all the images
    tsDLIterBD<CBlock> iter(pGif->blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	  // Update the XImage but not the pixmap, which will be
	  // updated when the image is displayed.
	    image->writeImageData();
	}
	++iter;
    }
}

void CGlobal::setLogicalScreenWidth(char *value)
{
  // Has to be two bytes, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xffff) val=0xffff;
    pGif->setLogicalScreenWidth(val);
    if(val != pGif->getLogicalScreenWidth()) {
	errMsg("CGlobal::setLogicalScreenWidth:\n"
	  "  Logical screen width must be %d to hold largest image\n",
	  pGif->getLogicalScreenWidth());
    }
    updateOptionsArea();
    displayData();
}

void CGlobal::setLogicalScreenHeight(char *value)
{
  // Has to be two bytes, unsigned
    long val=atol(value);
    if(val < 0) val=0;
    if(val > 0xffff) val=0xffff;
    pGif->setLogicalScreenHeight(val);
    if(val != pGif->getLogicalScreenHeight()) {
	errMsg("CGlobal::setLogicalScreenHeight:\n"
	  "  Logical screen height must be %d to hold largest image\n",
	  pGif->getLogicalScreenHeight());
    }
    updateOptionsArea();
    displayData();
}
